"""ChatGPT Plugin vector store."""

import os
from typing import Any, Dict, List, Optional

import requests
from requests.adapters import HTTPAdapter, Retry

from llama_index.schema import (
    BaseNode,
    MetadataMode,
    NodeRelationship,
    RelatedNodeInfo,
    TextNode,
)
from llama_index.vector_stores.types import (
    VectorStore,
    VectorStoreQuery,
    VectorStoreQueryResult,
)
from llama_index.utils import get_tqdm_iterable


def convert_docs_to_json(nodes: List[BaseNode]) -> List[Dict]:
    """Convert docs to JSON."""
    docs = []
    for node in nodes:
        # TODO: add information for other fields as well
        # fields taken from
        # https://rb.gy/nmac9u
        doc_dict = {
            "id": node.node_id,
            "text": node.get_content(metadata_mode=MetadataMode.NONE),
            # NOTE: this is the doc_id to reference document
            "source_id": node.ref_doc_id,
            # "url": "...",
            # "created_at": ...,
            # "author": "..."",
        }
        metadata = node.metadata
        if metadata is not None:
            if "source" in metadata:
                doc_dict["source"] = metadata["source"]
            if "source_id" in metadata:
                doc_dict["source_id"] = metadata["source_id"]
            if "url" in metadata:
                doc_dict["url"] = metadata["url"]
            if "created_at" in metadata:
                doc_dict["created_at"] = metadata["created_at"]
            if "author" in metadata:
                doc_dict["author"] = metadata["author"]

        docs.append(doc_dict)
    return docs


class ChatGPTRetrievalPluginClient(VectorStore):
    """ChatGPT Retrieval Plugin Client.

    In this client, we make use of the endpoints defined by ChatGPT.

    Args:
        endpoint_url (str): URL of the ChatGPT Retrieval Plugin.
        bearer_token (Optional[str]): Bearer token for the ChatGPT Retrieval Plugin.
        retries (Optional[Retry]): Retry object for the ChatGPT Retrieval Plugin.
        batch_size (int): Batch size for the ChatGPT Retrieval Plugin.
    """

    stores_text: bool = True
    is_embedding_query: bool = False

    def __init__(
        self,
        endpoint_url: str,
        bearer_token: Optional[str] = None,
        retries: Optional[Retry] = None,
        batch_size: int = 100,
        **kwargs: Any,
    ) -> None:
        """Initialize params."""
        self._endpoint_url = endpoint_url
        self._bearer_token = bearer_token or os.getenv("BEARER_TOKEN")
        self._retries = retries
        self._batch_size = batch_size

        self._s = requests.Session()
        self._s.mount("http://", HTTPAdapter(max_retries=self._retries))

    @property
    def client(self) -> None:
        """Get client."""
        return None

    def add(
        self,
        nodes: List[BaseNode],
    ) -> List[str]:
        """Add nodes to index."""
        headers = {"Authorization": f"Bearer {self._bearer_token}"}

        docs_to_upload = convert_docs_to_json(nodes)
        iterable_docs = get_tqdm_iterable(
            range(0, len(docs_to_upload), self._batch_size),
            show_progress=True,
            desc="Uploading documents",
        )
        for i in iterable_docs:
            i_end = min(i + self._batch_size, len(docs_to_upload))
            self._s.post(
                f"{self._endpoint_url}/upsert",
                headers=headers,
                json={"documents": docs_to_upload[i:i_end]},
            )

        return [result.node_id for result in nodes]

    def delete(self, ref_doc_id: str, **delete_kwargs: Any) -> None:
        """
        Delete nodes using with ref_doc_id.

        Args:
            ref_doc_id (str): The doc_id of the document to delete.

        """
        headers = {"Authorization": f"Bearer {self._bearer_token}"}
        self._s.post(
            f"{self._endpoint_url}/delete",
            headers=headers,
            json={"ids": [ref_doc_id]},
        )

    def query(
        self,
        query: VectorStoreQuery,
        **kwargs: Any,
    ) -> VectorStoreQueryResult:
        """Get nodes for response."""
        if query.filters is not None:
            raise ValueError("Metadata filters not implemented for ChatGPT Plugin yet.")

        if query.query_str is None:
            raise ValueError("query_str must be provided")
        headers = {"Authorization": f"Bearer {self._bearer_token}"}
        # TODO: add metadata filter
        queries = [{"query": query.query_str, "top_k": query.similarity_top_k}]
        res = requests.post(
            f"{self._endpoint_url}/query", headers=headers, json={"queries": queries}
        )

        nodes = []
        similarities = []
        ids = []
        for query_result in res.json()["results"]:
            for result in query_result["results"]:
                result_id = result["id"]
                result_txt = result["text"]
                result_score = result["score"]
                result_ref_doc_id = result["source_id"]
                node = TextNode(
                    id_=result_id,
                    text=result_txt,
                    relationships={
                        NodeRelationship.SOURCE: RelatedNodeInfo(
                            node_id=result_ref_doc_id
                        )
                    },
                )
                nodes.append(node)
                similarities.append(result_score)
                ids.append(result_id)

            # NOTE: there should only be one query
            break

        return VectorStoreQueryResult(nodes=nodes, similarities=similarities, ids=ids)



from typing import List, Any, Generator

import pytest
import asyncio
import os

from llama_index.schema import NodeRelationship, RelatedNodeInfo, TextNode
from llama_index.vector_stores import TimescaleVectorStore
from llama_index.vector_stores.timescalevector import IndexType
from llama_index.vector_stores.types import (
    VectorStoreQuery,
    MetadataFilters,
    ExactMatchFilter,
)

from datetime import timedelta, datetime

# from testing find install here https://github.com/timescale/python-vector/

TEST_SERVICE_URL = os.environ.get(
    "TEST_TIMESCALE_SERVICE_URL",
    "postgres://tsdbadmin:<password>@<id>.tsdb.cloud.timescale.com:<port>/tsdb?sslmode=require",
)
TEST_TABLE_NAME = "lorem_ipsum"

try:
    from timescale_vector import client  # noqa: F401

    cli = client.Sync(TEST_SERVICE_URL, TEST_TABLE_NAME, 1536)
    with cli.connect() as test_conn:
        pass

    cli.close()

    timescale_not_available = False
except (ImportError, Exception):
    timescale_not_available = True


@pytest.fixture(scope="session")
def conn() -> Any:
    import psycopg2

    conn_ = psycopg2.connect(TEST_SERVICE_URL)  # type: ignore
    return conn_


@pytest.fixture()
def db(conn: Any) -> Generator:
    conn.autocommit = True

    with conn.cursor() as c:
        c.execute(f"DROP TABLE IF EXISTS {TEST_TABLE_NAME}")
        conn.commit()
    yield
    with conn.cursor() as c:
        # c.execute(f"DROP TABLE IF EXISTS {TEST_TABLE_NAME}")
        conn.commit()


@pytest.fixture
def tvs(db: None) -> Any:
    tvs = TimescaleVectorStore.from_params(
        service_url=TEST_SERVICE_URL,
        table_name=TEST_TABLE_NAME,
    )

    yield tvs

    try:
        asyncio.get_event_loop().run_until_complete(tvs.close())
    except RuntimeError:
        asyncio.run(tvs.close())


@pytest.fixture
def tvs_tp(db: None) -> Any:
    tvs = TimescaleVectorStore.from_params(
        service_url=TEST_SERVICE_URL,
        table_name=TEST_TABLE_NAME,
        time_partition_interval=timedelta(hours=1),
    )

    yield tvs

    try:
        asyncio.get_event_loop().run_until_complete(tvs.close())
    except RuntimeError:
        asyncio.run(tvs.close())


@pytest.fixture(scope="session")
def node_embeddings() -> List[TextNode]:
    return [
        TextNode(
            text="lorem ipsum",
            id_="aaa",
            relationships={NodeRelationship.SOURCE: RelatedNodeInfo(node_id="aaa")},
            embedding=[1.0] * 1536,
        ),
        TextNode(
            text="dolor sit amet",
            id_="bbb",
            relationships={NodeRelationship.SOURCE: RelatedNodeInfo(node_id="bbb")},
            extra_info={"test_key": "test_value"},
            embedding=[0.1] * 1536,
        ),
    ]


@pytest.mark.skipif(
    timescale_not_available, reason="timescale vector store is not available"
)
@pytest.mark.asyncio
async def test_instance_creation(db: None) -> None:
    tvs = TimescaleVectorStore.from_params(
        service_url=TEST_SERVICE_URL,
        table_name=TEST_TABLE_NAME,
    )
    assert isinstance(tvs, TimescaleVectorStore)
    await tvs.close()


@pytest.mark.skipif(
    timescale_not_available, reason="timescale vector store is not available"
)
@pytest.mark.asyncio
@pytest.mark.parametrize("use_async", [(True), (False)])
async def test_add_to_db_and_query(
    tvs: TimescaleVectorStore, node_embeddings: List[TextNode], use_async: bool
) -> None:
    if use_async:
        await tvs.async_add(node_embeddings)
    else:
        tvs.add(node_embeddings)
    assert isinstance(tvs, TimescaleVectorStore)
    q = VectorStoreQuery(query_embedding=[1] * 1536, similarity_top_k=1)
    if use_async:
        res = await tvs.aquery(q)
    else:
        res = tvs.query(q)
    assert res.nodes
    assert len(res.nodes) == 1
    assert res.nodes[0].node_id == "aaa"


@pytest.mark.skipif(
    timescale_not_available, reason="timescale vector store is not available"
)
@pytest.mark.asyncio
@pytest.mark.parametrize("use_async", [(True), (False)])
async def test_add_to_db_and_query_with_metadata_filters(
    tvs: TimescaleVectorStore, node_embeddings: List[TextNode], use_async: bool
) -> None:
    if use_async:
        await tvs.async_add(node_embeddings)
    else:
        tvs.add(node_embeddings)
    assert isinstance(tvs, TimescaleVectorStore)
    filters = MetadataFilters(
        filters=[ExactMatchFilter(key="test_key", value="test_value")]
    )
    q = VectorStoreQuery(
        query_embedding=[0.5] * 1536, similarity_top_k=10, filters=filters
    )
    if use_async:
        res = await tvs.aquery(q)
    else:
        res = tvs.query(q)
    assert res.nodes
    assert len(res.nodes) == 1
    assert res.nodes[0].node_id == "bbb"
    assert res.ids is not None
    assert res.ids[0] == "bbb"


@pytest.mark.skipif(
    timescale_not_available, reason="timescale vector store is not available"
)
@pytest.mark.asyncio
@pytest.mark.parametrize("use_async", [(True), (False)])
async def test_async_add_to_db_query_and_delete(
    tvs: TimescaleVectorStore, node_embeddings: List[TextNode], use_async: bool
) -> None:
    if use_async:
        await tvs.async_add(node_embeddings)
    else:
        tvs.add(node_embeddings)
    assert isinstance(tvs, TimescaleVectorStore)

    q = VectorStoreQuery(query_embedding=[0.1] * 1536, similarity_top_k=1)

    if use_async:
        res = await tvs.aquery(q)
    else:
        res = tvs.query(q)
    assert res.nodes
    assert len(res.nodes) == 1
    assert res.nodes[0].node_id == "bbb"
    tvs.delete("bbb")

    if use_async:
        res = await tvs.aquery(q)
    else:
        res = tvs.query(q)
    assert res.nodes
    assert len(res.nodes) == 1
    assert res.nodes[0].node_id == "aaa"


@pytest.mark.skipif(
    timescale_not_available, reason="timescale vector store is not available"
)
def test_add_to_db_query_and_delete(
    tvs: TimescaleVectorStore, node_embeddings: List[TextNode]
) -> None:
    tvs.add(node_embeddings)
    assert isinstance(tvs, TimescaleVectorStore)

    q = VectorStoreQuery(query_embedding=[0.1] * 1536, similarity_top_k=1)
    res = tvs.query(q)
    assert res.nodes
    assert len(res.nodes) == 1
    assert res.nodes[0].node_id == "bbb"

    tvs.create_index()
    tvs.drop_index()

    tvs.create_index(IndexType.TIMESCALE_VECTOR, max_alpha=1.0, num_neighbors=50)
    tvs.drop_index()

    tvs.create_index(IndexType.PGVECTOR_IVFFLAT, num_lists=20, num_records=1000)
    tvs.drop_index()

    tvs.create_index(IndexType.PGVECTOR_HNSW, m=16, ef_construction=64)
    tvs.drop_index()


@pytest.mark.skipif(
    timescale_not_available, reason="timescale vector store is not available"
)
@pytest.mark.asyncio
@pytest.mark.parametrize("use_async", [(True), (False)])
async def test_time_partitioning_default_uuid(
    tvs_tp: TimescaleVectorStore, node_embeddings: List[TextNode], use_async: bool
) -> None:
    if use_async:
        await tvs_tp.async_add(node_embeddings)
    else:
        tvs_tp.add(node_embeddings)
    assert isinstance(tvs_tp, TimescaleVectorStore)

    q = VectorStoreQuery(query_embedding=[0.1] * 1536, similarity_top_k=1)

    if use_async:
        res = await tvs_tp.aquery(q)
    else:
        res = tvs_tp.query(q)
    assert res.nodes
    assert len(res.nodes) == 1
    assert res.nodes[0].node_id == "bbb"


@pytest.mark.skipif(
    timescale_not_available, reason="timescale vector store is not available"
)
@pytest.mark.asyncio
@pytest.mark.parametrize("use_async", [(True), (False)])
async def test_time_partitioning_explicit_uuid(
    tvs_tp: TimescaleVectorStore, node_embeddings: List[TextNode], use_async: bool
) -> None:
    t0 = datetime(2018, 1, 1, 0, 0, 0)
    t = t0
    for node in node_embeddings:
        node.id_ = str(client.uuid_from_time(t))
        t = t + timedelta(days=1)
    if use_async:
        await tvs_tp.async_add(node_embeddings)
    else:
        tvs_tp.add(node_embeddings)
    assert isinstance(tvs_tp, TimescaleVectorStore)

    q = VectorStoreQuery(query_embedding=[0.1] * 1536, similarity_top_k=1)

    if use_async:
        res = await tvs_tp.aquery(q)
    else:
        res = tvs_tp.query(q)
    assert res.nodes
    assert len(res.nodes) == 1
    assert res.nodes[0].node_id == node_embeddings[1].node_id
    assert res.ids is not None
    assert res.ids[0] != node_embeddings[1].node_id

    # make sure time filter works. This query should return only the first node
    q = VectorStoreQuery(query_embedding=[0.1] * 1536, similarity_top_k=4)
    if use_async:
        res = await tvs_tp.aquery(q, end_date=t0 + timedelta(minutes=1))
    else:
        res = tvs_tp.query(q, end_date=t0 + timedelta(minutes=1))

    assert res.nodes
    assert len(res.nodes) == 1

    # here the filter should return both nodes
    q = VectorStoreQuery(query_embedding=[0.1] * 1536, similarity_top_k=4)
    if use_async:
        res = await tvs_tp.aquery(q, end_date=t0 + timedelta(days=3))
    else:
        res = tvs_tp.query(q, end_date=t0 + timedelta(days=3))

    assert res.nodes
    assert len(res.nodes) == 2



from llama_index.text_splitter.markdown_splitter import MarkdownSplitter


def test_header_splits() -> None:
    markdown_splitter = MarkdownSplitter()

    splits = markdown_splitter.split_text(
        """# Header 1

Header 1 content

# Header 2
Header 2 content
    """
    )
    print(splits)
    assert len(splits) == 2


def test_non_header_splits() -> None:
    markdown_splitter = MarkdownSplitter()

    splits = markdown_splitter.split_text(
        """# Header 1

#Not a header
Also # not a header
    # Still not a header
    """
    )
    assert len(splits) == 1


def test_pre_header_content() -> None:
    markdown_splitter = MarkdownSplitter()

    splits = markdown_splitter.split_text(
        """
pre-header content

# Header 1
Content
## Sub-header
    """
    )
    assert len(splits) == 3



"""Markdown splitter."""
from typing import List, Optional
from llama_index.bridge.pydantic import Field

from llama_index.callbacks.base import CallbackManager
from llama_index.callbacks.schema import CBEventType, EventPayload
from llama_index.text_splitter.types import TextSplitter
import re


class MarkdownSplitter(TextSplitter):
    """Implementation of splitting text for Markdown files."""

    callback_manager: CallbackManager = Field(
        default_factory=CallbackManager, exclude=True
    )

    def __init__(
        self,
        callback_manager: Optional[CallbackManager] = None,
    ):
        callback_manager = callback_manager or CallbackManager([])
        super().__init__(
            callback_manager=callback_manager,
        )

    @classmethod
    def class_name(cls) -> str:
        """Get class name."""
        return "MarkdownSplitter"

    def split_text(self, text: str) -> List[str]:
        if text == "":
            return []

        with self.callback_manager.event(
            CBEventType.CHUNKING, payload={EventPayload.CHUNKS: [text]}
        ):
            return self._split_text(text)

    def _split_text(self, text: str) -> List[str]:
        markdown_sections = []
        lines = text.split("\n")

        current_section = ""

        for line in lines:
            header_match = re.match(r"^(#+)\s(.*)", line)
            if header_match:
                if current_section != "":
                    markdown_sections.append(current_section.strip())
                current_section = f"{header_match.group(2)}\n"
            else:
                current_section += line + "\n"

        markdown_sections.append(current_section.strip())

        return markdown_sections
